import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from artemis7.stats.binning import *

@staticmethod
def isdiscrete(x):
    if isinstance(x, pd.Series):
        x = x.to_numpy()
    elif isinstance(x, list):
        x = np.array(x)
    
    nuniq = len(np.unique(x))
    nbin = sturges(x)
    if nuniq < nbin:
        return True
    else:
        False

class MVSampler():
    def __init__(self, reference, inference, target_cols):
        """
        reference : pd.DataFrame.
        inference : pd.DataFrame.
        target_cols : array-like.
        """
        self._pdf = {}
        self._cnt = np.zeros(reference.shape[0])
        self.reference = reference.copy()
        del reference
        self.inference = inference.copy()
        del inference
        self.target_cols = target_cols

        bins = sturges(self.inference.shape[0])
        for col in self.target_cols:
            inf = self.inference[col].to_numpy()
            if isdiscrete(inf): #해당 column이 discrete 하다면
                b, p = np.unique(inf, return_counts=True)
                p, b = p[np.argsort(b)], np.sort(b)

            else: # continuous하다면
                p, bin_edge = np.histogram(self.inference[col].values,
                                            bins=np.min([bins, len(self.inference[col].unique())]),
                                            # bins = bins,
                                            density=True)
                b, bin_width = (bin_edge[1:] + bin_edge[:-1])/2, (bin_edge[1]-bin_edge[0])/2
                
                b, p = np.insert(b, 0, self.inference[col].min()-bin_width), np.insert(p, 0, 0)
                b, p = np.append(b, self.inference[col].max()+bin_width), np.append(p, 0)
            
            f = interp1d(b, p, kind="linear")
            self._pdf[col] = (p, b, f)
            lcond = self.reference[col].values>=b.min()
            rcond = self.reference[col].values<=b.max()
            self._cnt += (lcond*rcond*1)

    def build(self, row_min = -1, outlier = 0):
        """
        row_min : int. sampling에 사용될 reference의 최소 row 갯수
        outlier : int. ref/inf 의 target_col 분포를 비교할 때, 고려하지 않을 column의 갯수
        모든 target_col이 ref/inf 분포가 겹치지 않을 수 있기 때문에 일정 row 갯수가 확보될 때 까지 증가
        """
        if row_min  < 0:
            row_min = int(self.inference.shape[0] * 0.1)

        nsize = len(np.where(self._cnt>=len(self.target_cols)-outlier)[0])
        target_cols = self.target_cols
        while (nsize <= row_min) and (outlier < len(self.target_cols)):
            outlier += 1
            target_cols = self.target_cols[:len(self.target_cols)-outlier]
            self._cnt = np.zeros(self.reference.shape[0])
            for col in target_cols:
                b = self._pdf[col][1]
                lcond = self.reference[col].values>=b.min()
                rcond = self.reference[col].values<=b.max()
                self._cnt += (lcond*rcond*1)
            nsize = len(np.where(self._cnt>=len(target_cols))[0])
        self._outlier = outlier

        self.log_likelihood = np.zeros(self.reference.shape[0])
        for i, idx in enumerate(self.reference.index.to_numpy()):
            if self._cnt[i]>=len(target_cols):
                p = []
                for col in target_cols:
                    _pdf = self._pdf[col]
                    if self.reference.at[idx, col]<=_pdf[1].min() or self.reference.at[idx, col]>=_pdf[1].max():
                        p.append(0)
                    else:
                        p.append(self._pdf[col][2](self.reference.at[idx, col]))
                p = np.array(p)
                self.log_likelihood[i] = np.sum(np.log10(p[np.where(p!=0)[0]]))
                # np.power(10, np.sum(np.log10(p[np.where(p!=0)[0]])))

    def sample(self, n=-1, seed=None, return_index=False):
        """
        n : int. sample의 크기
        return_index : bool. True일 경우 sample의 index를 array로 return, False일 경우 sample을 pd.DataFrame으로 return. 기본은 False
        return : pd.DataFrame / array-like
        """
        n = n if n>0 else self.inference.shape[0]//10
        prob = np.power(10, self.log_likelihood)
        zeros = np.where(self.log_likelihood==0)[0]
        prob[zeros] = 0
        prob /= np.sum(prob)

        rng = np.random.default_rng(seed)
        idx = rng.choice(np.arange(0, self.reference.shape[0]), replace=True, p=prob, size=n)
        if return_index:
            return idx
        return self.reference.iloc[idx, :]


import numpy as np

@staticmethod
def __getBin(rule, x, axis):
    result = None
    if axis==None:
        result = rule(x)
    elif axis==0:
        r = np.array([rule(x[i, :]) for i in range(x.shape[0])])
        result = r.reshape(-1, 1)
    elif axis==1:
        r = np.array([rule(x[:, i]) for i in range(x.shape[1])])
        result = r.reshape(1, -1)
    else:
        raise np.AxisError("axis %s is out of bounds for array of dimension %d"%(axis, x.ndim))
    return result

def sturges(x, axis=None):
    def __sturges(x):
        if np.ndim(x)==0:
            n = x
        else:
            n = len(x)
        return int(1 + np.ceil(np.log2(n)))
    return __getBin(__sturges, x, axis)

def doane(x, axis=None):
    def __doane(x):
        n = len(x)
        g1 = np.sum( (x - np.mean(x))**3 )
        g1 /= (np.std(x)**2)**(3/2)
        sig_g1 = np.sqrt(6*(n - 2)/(n+1)/(n+3))
        return  int(1 + np.log2(n) + np.log2(1+ abs(g1)/sig_g1))
    return __getBin(__doane, x, axis)

def scott(x, axis=None):
    def __scott(x):
        n = len(x)
        h = 3.49 * np.std(x) / (n**(1/3))
        return int(max(x)-min(x)/h)
    return __getBin(__scott, x, axis)

def freedman_diaconis(x, axis=None):
    def __freedman_diaconis(x):
        n = len(x)
        iqr = np.percentile(x, 75) - np.percentile(x, 25)
        h = 2*iqr/(n**(1/3))
        return int(max(x)-min(x)/h)
    return __getBin(__freedman_diaconis, x, axis)



import numpy as np

class PRS():
    def __init__(self, X, bins, seed=False):
        if seed:
            np.random.seed(seed)
        self.X = X
        self._xrange = (np.min(X), np.max(X))
        self._bins = bins
        self._wi, self._x_edge = np.histogram(self.X, bins = self._bins, range=self._xrange, density=True)
        self._wi_csum = np.cumsum(self._wi)
    
    def sample(self, data, n=100, resampling=False, index=False):
        """
        data : Target population
        n : The number of samples. Default is 100
        resampling : resample or not. Default is False
        index : return index rather value. Default is False
        """

        y, _ = np.histogram(data, bins=self._bins, range=self._xrange, density=True)
        X_max_bin = np.where(self._wi == np.max(self._wi))[0]
        n_limit = int(len(self.X) * y[X_max_bin] / self._wi[X_max_bin]*0.9)
        n = np.min([n, n_limit])

        result, cnt = np.zeros(n), 0
        while cnt<n:
            Y = np.random.random() * self._wi_csum[-1]
            idx = 0

            while self._wi_csum[idx]<Y:
                idx += 1
            sample_area = np.where((data >= self._x_edge[idx]) & (data<self._x_edge[idx + 1]))[0]

            if len(sample_area):
                sampled_idx = np.random.choice(sample_area)
                if index:
                    result[cnt] = sampled_idx.astype(np.int64)
                else:
                    result[cnt] = data[sampled_idx]
                cnt += 1

                if not resampling:
                    data = np.delete(data, sampled_idx)
        return result
    
    def generate(self, n=100, mode="uniform"):
        """
        Generate the random numbers following the given distribution
        n : The number of samples. Default is 100
        mode : The distribution for the sampling
        """
        result, cnt = np.zeros(n), 0
        bin_width = self._x_edge[1] - self._x_edge[0]

        while cnt < n:
            Y = np.random.random() * self._wi_csum[-1]
            idx = 0

            while self._wi_csum[idx] < Y:
                idx += 1
            
            if mode == "uniform":
                result[cnt] = np.random.rand() * (bin_width / 2) + (self._x_edge[idx -1] + self._x_edge[idx]) / 2
            elif mode == "gaussian":
                result[cnt] = np.random.randn() * (bin_width / 2) + (self._x_edge[idx -1] + self._x_edge[idx]) / 2
            cnt += 1
        return result



import numpy as np

def population_stability_index(expected, actual, buckettype='bins', buckets=10, axis=0):
    def psi(expected_array, actual_array, buckets):
        
        def scale_range(input, min, max):
            input += -(np.min(input))
            input /= np.max(input) / (max - min)
            input += min
            return input

        breakpoints = np.arange(0, buckets + 1) / (buckets) * 100

        if buckettype == 'bins':
            breakpoints = scale_range(breakpoints, np.min(
                expected_array), np.max(expected_array))
        elif buckettype == 'quantiles':
            breakpoints = np.stack(
                [np.percentile(expected_array, b) for b in breakpoints])

        expected_percents = np.histogram(expected_array, breakpoints)[0] / len(expected_array)
        actual_percents = np.histogram(actual_array, breakpoints)[0] / len(actual_array)

        def sub_psi(e_perc, a_perc):
            if a_perc == 0:
                a_perc = 0.0001
            if e_perc == 0:
                e_perc = 0.0001

            value = (e_perc - a_perc) * np.log(e_perc / a_perc)
            return(value)

        psi_value = np.sum(sub_psi(expected_percents[i], actual_percents[i]) for i in range(
            0, len(expected_percents)))

        return(psi_value)

    if len(expected.shape) == 1:
        psi_values = np.empty(len(expected.shape))
    else:
        psi_values = np.empty(expected.shape[axis])

    for i in range(0, len(psi_values)):
        if len(psi_values) == 1:
            psi_values = psi(expected, actual, buckets)
        elif axis == 0:
            psi_values[i] = psi(expected[:, i], actual[:, i], buckets)
        elif axis == 1:
            psi_values[i] = psi(expected[i, :], actual[i, :], buckets)

    return psi_values


import numpy as np
import time




class progressbar():
    def __init__(self, maxV, jobname=""):
        self.maxV = maxV
        self._current_percent = 1
        self.t0 = time.time()
        self._banner_size = 10
        self.jobname = "."*(max([self._banner_size-len(jobname), 1]))+jobname
        
    def start(self):
        print(self._pout(0), end="")
        self.t0 = time.time()
    
    def _ETA(self, progress):
        t_now = time.time()
        t_passed = t_now - self.t0
        eta = t_passed * ((100-progress)/(progress+1e-6))
        h = eta//3600
        eta -= h*3600
        m = eta//60
        s = eta%60
        if h>0:
            return "ETA %dH:%02dm"%(h, m)
        return "ETA %02dm:%02ds"%(m, s)

    def _pout(self, percent, eta_disp=True):
        bk = 4-len(str(int(percent)))
        bar = "| "  + "█"*int(percent//2) + " "*int(50-percent//2)
        percentage = " "*bk +"%d%%"%percent+ "|"
        if eta_disp:
            eta = self._ETA(percent)
            return  bar + percentage + eta + self.jobname
        return bar + percentage + self.jobname

    def update(self, c):
        percent = c/self.maxV*100
        if abs(self.maxV-c) <= 1:
            self.__del__()
        if np.ceil(percent)>=self._current_percent:
            print("\r"+self._pout(percent), end="")
            self._current_percent = percent
    
    def __del__(self):
        print("\r"+self._pout(100, eta_disp=True), end="")



import time
n = 50
pb = progressbar(n, "hi")
pb.start()
for i in range(n):
    pb.update(i)
    time.sleep(0.05)




import matplotlib.pyplot as plt
import matplotlib.font_manager as fm
import matplotlib.cm as cm
from matplotlib.colors import Normalize
from matplotlib.ticker import MultipleLocator
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

import numpy as np
import sys
sys.path.append(sys.path[0]+"/..")
from artemis7.stats.binning import *

## init ##
plt.rcParams["font.family"] = "Times New Roman"
plt.rcParams["xtick.labelsize"] = 12
plt.rcParams["ytick.labelsize"] = 12


def plot(ax, xy, **kwargs):
    m = cm.ScalarMappable(norm = Normalize(vmin = 0, vmax = len(xy)), cmap=plt.get_cmap("viridis"))
    xr, yr = (np.inf, -np.inf), (np.inf, -np.inf)
    for i, line in enumerate(xy):
        args = {k:v[i] for k, v in kwargs.items()}
        ax.plot(*line, color="w", linewidth=3)
        ax.plot(*line, **args, color=m.to_rgba(i))
        xr = (np.min([xr[0], line[0].min()]), np.max([xr[1], line[0].max()]))
        yr = (np.min([yr[0], line[1].min()]), np.max([yr[1], line[1].max()]))
    ax.grid()
    ax.set_xlim(xr[0], xr[1])
    ax.set_ylim(yr[0], yr[1])
    return

def colored_scatter(ax, x, y, c):
    m = cm.ScalarMappable(norm = Normalize(vmin = np.min(c), vmax = np.max(c)), cmap=plt.get_cmap("viridis"))
    ax.scatter(x, y, facecolor=m.to_rgba(c), edgecolor="k", alpha=0.5)
    axins = inset_axes(
        ax,
        width = "20%",
        height= "3%",
        loc=3
    )
    plt.colorbar(m, cax = axins, orientation = "horizontal", ticks=[np.min(c), np.max(c)])
    axins.xaxis.set_ticks_position("top")

    ax.grid()
    ax.set_xlim(np.min(x), np.max(x))
    ax.set_ylim(np.min(y), np.max(y))
    return


def number_density(ax, x, y, **kwargs):
    bin = sturges(x)
    sigma_ratio = {0.954:r"$2\sigma$", 0.866:r"$1.5\sigma$", 0.683:r"$1\sigma$"}
    img, xe, ye = np.histogram2d(x, y, bins=bin)
    img_sorted_1d = np.sort(img.reshape(-1))[::-1]
    img_cs = np.cumsum(img_sorted_1d)/np.sum(img_sorted_1d)
    ext = [xe[0], xe[-1], ye[0], ye[-1]]

    lv = np.array([img_sorted_1d[np.where(img_cs > level)[0][0]] for level in sigma_ratio.keys()])
    
    ax.contourf(img.T, extent = ext, cmap=plt.get_cmap("cubehelix_r"), alpha=0.7, vmin = np.percentile(img_sorted_1d, 68), vmax= img_sorted_1d.max(), origin="lower")
    cs = ax.contour(img.T, extent = ext, levels = lv, colors='k', origin="lower")
    plt.clabel(cs, levels=lv, fontsize=10, fmt={l:v for (l, v) in zip(lv, sigma_ratio.values())})
    ax.scatter(np.mean(x), np.mean(y), marker="x", color="w", s= 50)
    ax.set_xlim(np.min(x), np.max(x))
    ax.set_ylim(np.min(y), np.max(y))
    ax.grid()
    return

def number_density_histxy(axc, axu, axr, x, y):
    bin = sturges(x)
    axc = number_density(axc, x, y)

    axu.hist(x, bins= bin, range=(np.min(x), np.max(x)), histtype="step", color="k")
    axu.set_xlim(np.min(x), np.max(x))
    axu.grid(axis="x")
    axu.xaxis.tick_top()
    axu.xaxis.set_label_position('top')

    axr.hist(y, bins= bin, range=(np.min(y), np.max(y)), histtype="step", orientation="horizontal", color="k")
    axr.set_ylim(np.min(y), np.max(y))
    axr.grid(axis="y")
    axr.yaxis.tick_right()
    axr.yaxis.set_label_position('right')
    return


def bubble_diagram(ax, x, y, w, **kwargs):
    minS, maxS = 5, 200
    scale = (w - np.min(w))/(np.max(w)-np.min(w))
    scale *= (maxS- minS)
    scale += minS

    ax.scatter(x, y, s= scale, facecolor="None", edgecolor="k", **kwargs)
    ax.grid()
    return

def legend(ax, **kwargs):
    legend_kwargs = {
                'fontsize' : 12,
                'framealpha' : 0.9,
                'facecolor' : "w",
                'edgecolor' : 'k',
                'shadow' : False,
                'labelspacing' : 0.1,
            }
    ax.legend(**kwargs, **legend_kwargs)

def set_background(ax, **kwargs):
    ax.patch.set_facecolor('skyblue')
    ax.patch.set_alpha(0.05)
    return


def set_minorticks(ax, axis="both", **kwargs):
    def xaxis(ax):
        majorticks = ax.get_xticks()
        minorbin = abs(majorticks[1]-majorticks[0])/5
        ax.xaxis.set_minor_locator(MultipleLocator(minorbin))
        return
    def yaxis(ax):
        majorticks = ax.get_yticks()
        minorbin = abs(majorticks[1]-majorticks[0])/5
        ax.yaxis.set_minor_locator(MultipleLocator(minorbin))
    if axis == "x":
        xaxis(ax)
    elif axis == "y":
        yaxis(ax)
    else:
        xaxis(ax)
        yaxis(ax)
    return

def adjust_tickposition(ax, loc, **kwargs):
    """
        261
        795
        384
    """
    if loc == 1 or loc == 2 or loc == 6:
        ax.xaxis.tick_top()
        ax.xaxis.set_label_position('top')
    if loc == 1 or loc == 5 or loc == 4:
        ax.yaxis.tick_right()
        ax.yaxis.set_label_position('right')
    return


import numpy as np
import pandas as pd
from numpy.random import choice, rand, gamma, randint
from astropy.time import Time
from datetime import datetime
import json
from artemis7.stats.binning import *

"""
Virtual drift generator (VDG)


1. Initialization

Between (time_range[0]~time_range[1]), Drifts take place at 'drfit_time_sequence'.
'time_range' and 'drift_time_sequence' support float (mjd) and string (datetime64, yyyy-mm-dd)

CDG only support continuous columns and 2type drift mode (base<->drift) (2023-01-06).
CDG can support continous and discrete columns (2023-01-12)
CDG supports regression prob type (2023-01-19)

2. Generation
X (Data set) is randomly sampled with Gamma distribution (https://en.wikipedia.org/wiki/Gamma_distribution).
Gamma distribution has two parameters (k, theta). The parameters follow normal distribution in this class.

generate() methods supports 2 types of drift modes [real/virtual] and 3types of drift patterns [sudden/incremental/gradual].
In case of drift mode, "real" is the drift in p(y|X) which is the case that X doesn't change but label y change only.
For "virtual", p(y|X) doesn't change but the distribution of X changes.
you can control the strength of the drift with parameter strength. 'strength' works linearly in "real" and logarithmic in "virtual".
Drift pattern depends on the time interval of [drift_start, drift_end].
"sudden" make drift with time interval=0.
"Incremental" and "gradual" have same time interval of [0.3*dT_bef, 0.3*dT_aft].
The difference between them is that "incremental" has gentle drift in time interval and "gradual" has abrupting dirft.



"""

class VirtualDriftGenerator():
    def __init__(self, time_range, drift_time_sequence, prob_type, n_cont, n_disc, **kwargs):
        """
        time_range : array-like. 데이터셋의 시작시간과 끝시간. (start, end)형태로 입력.
        drift_time_sequence : array-like. drift가 작동하는 시간.
        prob_type : [classification / regression] y의 type
        n_cont : 연속데이터 컬럼의 갯수
        n_disc : 불연속데이터 컬럼의 갯수
        size : Integer. data point 갯수
        dt : Float. data point의 간격
        """

        if not time_range[0].isdigit():
            time_range = np.sort(Time(time_range).mjd)

        
        if not drift_time_sequence[0].isdigit():
            drift_time_sequence = np.sort(Time(drift_time_sequence).mjd)

        keys = kwargs.keys()

        if "size" in keys:
            self._size = kwargs["size"]
            self.time = np.linspace(time_range[0], time_range[1], self._size)
        elif "dt" in keys:
            self.time = np.arange(time_range[0], time_range[1], kwargs["dt"])
            self._size = len(self.time)
        else:
            raise AssertionError("Absence in time binning")
        
        self._drift_weight = np.zeros(len(self.time))
        
        self._ncol = (n_cont, n_disc)
        self.n_att = kwargs["n_att"] if "n_att" in keys else sum(self._ncol)//2

        if sum(self._ncol)<=0:
            raise ValueError("At least 1 column should be exist")

        tol = 1e-6
        self._drift_sequence = np.sort(np.concatenate((drift_time_sequence, [time_range[1]+tol , time_range[0]-tol]), axis=0))
        self.prob_type = prob_type
        yfunction = {
            "classification" : self._classfication,
            "regression" : self._regression}
        self._label_function = yfunction[self.prob_type]
        self.noise = 0
        self.attribute_columns = []
        self.df = None
        self.df_info = {
            "file" : None,
            "prob_type" : self.prob_type,
            "date_generated" : None,
            "time_range" : time_range,
            "drift_time" : drift_time_sequence,
            "n_cont" : n_cont,
            "n_disc" : n_disc,
            "drift": None,
            "drift_type" : None,
            "strength" : 0,
            "noise" : 0,
            "att_cols": self.attribute_columns
        }

    @staticmethod
    def _classfication(X, att_cols, threshold):
        X = X[:, att_cols]
        sumX = np.sum(X, axis=1)
        return (sumX>threshold).astype(np.int32)

    @staticmethod
    def _regression(X, att_cols, threshold):
        X = X[:, att_cols]
        sumX = np.sum(X, axis=1)
        return sumX+threshold
    
    @staticmethod
    def _discrete_sampling(sample, pdf, size):
        if len(sample) != len(pdf):
            raise ValueError("The sample and PDF have different size")
        pdf /= np.sum(pdf)
        cdf = np.cumsum(pdf)
        cdf = np.insert(cdf, 0, 0)

        data = np.zeros(size)
        dums = np.random.random(size)

        for idx in range(size):
            pidx = 0
            while cdf[pidx]<dums[idx]:
                pidx += 1
            pidx -= 1
            data[idx] = sample[pidx]
        data = data.astype(int)
        return data[0] if size==1 else data

    @staticmethod
    def hist2d(df0, df1, dropcol):
        v0 = df0.drop(columns=dropcol).to_numpy()
        v1 = df1.drop(columns=dropcol).to_numpy()
        
        b0 = sturges(v0[:, 0])
        b1 = sturges(v1[:, 0])
        bins = np.max([b0, b1])

        xr = (np.min([v0[:, 0].min(), v1[:, 0].min()]), np.max([v0[:, 0].max(), v1[:, 0].max()]))
        a0, _ = np.histogram(v0[:, 0], bins=bins, range=xr, density=True)
        a1, _ = np.histogram(v1[:, 0], bins=bins, range=xr, density=True)
        for i in range(1, len(v0[0, :])):
            xr = (np.min([v0[:, i].min(), v1[:, i].min()]), np.max([v0[:, i].max(), v1[:, i].max()]))
            h0, _ = np.histogram(v0[:, i], bins=bins, range=xr, density=True)
            h1, _ = np.histogram(v1[:, i], bins=bins, range=xr, density=True)
            a0 = np.vstack((a0, h0))
            a1 = np.vstack((a1, h1))
        return np.sum(np.abs(a0-a1))/bins/len(a0)
    
    @staticmethod
    def read_info(path, file_name):
        with open(path+"%s.json"%file_name, "r") as f:
            file_info = json.load(f)
        return file_info
    
    def _sanity_check(self, df):
        dropcol = ["y", "time"]
        not_ok = False
        d0 = df[df.time<=(self._drift_sequence[0]+ self._drift_sequence[1])/2]
        d1 = df[(df.time>(self._drift_sequence[0]+self._drift_sequence[1])/2) &(df.time<=self._drift_sequence[1])]
        score0 = self.hist2d(d0, d1, dropcol)
        on_drift = False
        for ts, te in zip(self._drift_sequence[:-1], self._drift_sequence[1:]):
            d_now = df[(df.time>ts) & (df.time<te)]
            score = self.hist2d(d0, d_now, dropcol)

            not_ok = on_drift and (score0>score)
            if not_ok:
                return True        
        return not_ok

    def _add_noise_cont(self, base_col):
        pos = np.random.random(self._size)
        pos = np.where(pos<=self.noise)[0]
        base_col[pos] = base_col[pos]*np.random.normal(loc=1, size=len(pos))
        return base_col
    
    def _add_noise_disc(self, base_col):
        pos = np.random.random(self._size)
        pos = np.where(pos<=self.noise)[0]
        uniq = np.unique(base_col)
        for p in pos:
            base_col[p] = np.random.choice(uniq)
        return base_col


    def _real_sudden(self, base, thresholds, att_cols):
        y = np.array([])
        for idx, (ts, te) in enumerate(zip(self._drift_sequence[:-1], self._drift_sequence[1:])):
            X = base[(base.time>ts) & (base.time<=te)].drop(columns="time").to_numpy()
            threshold = thresholds[idx%2]
            yadd = self._label_function(X = X,
                                    att_cols = att_cols,
                                    threshold = threshold)
            y = np.concatenate((y,yadd),axis=0)
            self._drift_weight[np.where((self.time>ts)&(self.time<=te))[0]] = idx%2
        base["drift"] = self._drift_weight
        base["y"] = y
        return base
    
    def _real_incremental(self, base, thresholds, att_cols):
        y = np.array([])
        on_drift = False
        for idx, (ts, te) in enumerate(zip(self._drift_sequence[:-1], self._drift_sequence[1:])):
            X = base[(base.time>ts) & (base.time<=te)].drop(columns="time").to_numpy()

            threshold_bef = thresholds[0] if on_drift else thresholds[1]
            threshold_now = thresholds[1] if on_drift else thresholds[0]

            cur_interval = np.where((self.time>ts) & (self.time<=te))[0]
            itv_len = len(cur_interval)
            l = int(0.3*itv_len)
            pre_pivot = cur_interval[int(itv_len*0.3)]
            suf_pivot = cur_interval[int(itv_len*0.7)]

            for c, i in enumerate(cur_interval):
                if i<pre_pivot and ts != self._drift_sequence[0]:
                    bef = (l-c)/(2*l)
                elif i>suf_pivot and te != self._drift_sequence[-1]:
                    bef = (l-(itv_len-c))/(2*l)
                else:
                    bef = 0
                X_sub = X[c, :].reshape(1, -1)
                threshold = threshold_bef*bef + threshold_now*(1-bef)
                yadd = self._label_function(X = X_sub,
                                    att_cols = att_cols,
                                    threshold = threshold)
                y = np.concatenate((y,yadd),axis=0)
                self._drift_weight[i] = 1-bef if on_drift else bef
            on_drift = not on_drift
        base["drift"] = self._drift_weight
        base["y"] = y
        return base
    
    def _real_gradual(self, base, thresholds, att_cols):
        y = np.array([])
        on_drift = False
        for idx, (ts, te) in enumerate(zip(self._drift_sequence[:-1], self._drift_sequence[1:])):
            X = base[(base.time>ts) & (base.time<=te)].drop(columns="time").to_numpy()

            threshold_bef = thresholds[0] if on_drift else thresholds[1]
            threshold_now = thresholds[1] if on_drift else thresholds[0]

            cur_interval = np.where((self.time>ts) & (self.time<=te))[0]
            itv_len = len(cur_interval)
            l = int(0.3*itv_len)
            pre_pivot = cur_interval[int(itv_len*0.3)]
            suf_pivot = cur_interval[int(itv_len*0.7)]

            for c, i in enumerate(cur_interval):
                if i<pre_pivot and ts != self._drift_sequence[0]:
                    bef = (l-c)/(2*l)
                elif i>suf_pivot and te != self._drift_sequence[-1]:
                    bef = (l-(itv_len-c))/(2*l)
                else:
                    bef = 0
                X_sub = X[c, :].reshape(1, -1)
                cho = np.random.random()<=bef
                threshold = threshold_bef if cho else threshold_now
                yadd = self._label_function(X = X_sub,
                                    att_cols = att_cols,
                                    threshold = threshold)
                y = np.concatenate((y,yadd),axis=0)
                self._drift_weight[i] = 1-int(cho) if on_drift else int(cho)
            on_drift = not on_drift
        base["drift"] = self._drift_weight
        base["y"] = y
        return base
    
    def _generate_real(self, ncol, strength, drift_type="sudden"):
        n_cont, n_disc = map(int, ncol)
        base_coeff = rand(n_cont, 2)*10 #G[[k, theta]*n_cont]

        nsamples = randint(6, 30, n_disc)
        samples = [choice(np.arange(nsamples[i]), nsamples[i]//2, replace=False) for i in range(n_disc)]
        base_pdf = [rand(len(samples[i])) for i in range(n_disc)]

        base = {"time" : self.time}
        attribute_cols = choice(np.arange(np.sum(ncol)), self.n_att, replace=False)

        self.attribute_columns = np.sort(attribute_cols)
        dumx = np.zeros(self._size) #rename
        for idx, (k, theta) in enumerate(base_coeff):
            gamma_dist = gamma(k, theta, self._size)
            base["C%02d"%idx] = gamma_dist
            if idx in attribute_cols:
                dumx += gamma_dist

        for idx in range(n_disc):
            sample = self._discrete_sampling(samples[idx], base_pdf[idx], self._size)
            base["D%02d"%idx] = sample
            if int(idx+n_cont) in attribute_cols:
                dumx += sample

        gamma_dist /= len(attribute_cols)
        att_mean, att_std = np.mean(dumx), np.std(dumx)
        thresholds = [att_mean+att_std*strength/20, att_mean-att_std*strength/20]
        base = pd.DataFrame(base)

        if drift_type == "sudden":
            df = self._real_sudden(base = base,
            thresholds=thresholds,
            att_cols=attribute_cols)
            
        elif drift_type == "incremental":
            df = self._real_incremental(base = base,
            thresholds=thresholds,
            att_cols=attribute_cols)
        
        elif drift_type == "gradual":
            df = self._real_gradual(base = base,
            thresholds=thresholds,
            att_cols=attribute_cols)
        else:
            raise KeyError("There is no such type of drift [%s]"%drift_type)
        return df
    
    def _virtual_sudden(self, base, drift_coeff, samples, base_pdf, drift_pdf):
        on_drift=False
        for idx, (ts, te) in enumerate(zip(self._drift_sequence[:-1], self._drift_sequence[1:])):
            cond = (base.time>ts) & (base.time<=te)
            sub_size = np.sum(cond.to_numpy())
            if on_drift:
                for i, (k, theta) in enumerate(drift_coeff):
                    temp_g = gamma(k, theta, sub_size)
                    base.loc[cond, "C%02d"%i] = temp_g
                
                for i in range(len(samples)):
                    sample, pdf = samples[i], drift_pdf[i]
                    temp_d = self._discrete_sampling(sample, pdf, sub_size)
                    base.loc[cond, "D%02d"%i] = temp_d

            self._drift_weight[np.where((self.time>ts)&(self.time<=te))[0]] = idx%2
            on_drift = not on_drift
        return base
    
    def _virtual_incremental(self, base, base_coeff, drift_coeff, samples, base_pdf, drift_pdf):
        on_drift=False
        for idx, (ts, te) in enumerate(zip(self._drift_sequence[:-1], self._drift_sequence[1:])):
            coeff_now = base_coeff if on_drift else drift_coeff
            coeff_bef = drift_coeff if on_drift else base_coeff

            pdf_now = base_pdf if on_drift else drift_pdf
            pdf_bef = drift_pdf if on_drift else base_pdf

            cur_interval = np.where((self.time>ts) & (self.time<=te))[0]
            itv_len = len(cur_interval)
            l = int(0.3*itv_len)
            pre_pivot = cur_interval[int(itv_len*0.3)]
            suf_pivot = cur_interval[int(itv_len*0.7)]

            for c, i in enumerate(cur_interval):
                if i<pre_pivot and ts != self._drift_sequence[0]:
                    bef = (l-c)/(2*l)
                elif i>suf_pivot and te != self._drift_sequence[-1]:
                    bef = (l-(itv_len-c))/(2*l)
                else:
                    bef = 0                    
                coeffs = coeff_bef*bef + coeff_now*(1-bef)
                X_sub = np.array([gamma(k, theta) for k, theta in coeffs])
                for j, x in enumerate(X_sub):
                    base.loc[i, "C%02d"%j] = x

                pdfs_ = [pdf_bef[i] * bef + pdf_now[i] * (1-bef) for i in range(len(pdf_bef))]
                X_sub = np.array([
                    self._discrete_sampling(sample, pdf, 1) for sample, pdf in zip(samples, pdfs_)
                    ])
                for j, x in enumerate(X_sub):
                    base.loc[i, "D%02d"%j] = x

                self._drift_weight[i] = 1-bef if on_drift else bef
                
            on_drift = not on_drift
        return base
    
    def _virtual_gradual(self, base, base_coeff, drift_coeff, samples, base_pdf, drift_pdf):
        on_drift=False
        for idx, (ts, te) in enumerate(zip(self._drift_sequence[:-1], self._drift_sequence[1:])):
            coeff_now = base_coeff if on_drift else drift_coeff
            coeff_bef = drift_coeff if on_drift else base_coeff

            pdf_now = base_pdf if on_drift else drift_pdf
            pdf_bef = drift_pdf if on_drift else base_pdf

            cur_interval = np.where((self.time>ts) & (self.time<=te))[0]
            itv_len = len(cur_interval)
            l = int(0.3*itv_len)
            pre_pivot = cur_interval[int(itv_len*0.3)]
            suf_pivot = cur_interval[int(itv_len*0.7)]

            for c, i in enumerate(cur_interval):
                if i<pre_pivot and ts != self._drift_sequence[0]:
                    bef = (l-c)/(2*l)
                elif i>suf_pivot and te != self._drift_sequence[-1]:
                    bef = (l-(itv_len-c))/(2*l)
                else:
                    bef = 0
                
                cho = np.random.random()<=bef
                coeffs = coeff_bef if cho else coeff_now
                X_sub = np.array([gamma(k, theta) for k, theta in coeffs])
                for j, x in enumerate(X_sub):
                    base.loc[i, "C%02d"%j] = x

                pdfs_ = [pdf_bef[i]*bef + pdf_now[i]*(1-bef) for i in range(len(pdf_bef))]
                X_sub = np.array([
                    self._discrete_sampling(sample, pdf, 1) for sample, pdf in zip(samples, pdfs_)
                    ])
                for j, x in enumerate(X_sub):
                    base.loc[i, "D%02d"%j] = x

                self._drift_weight[i] = 1-int(cho) if on_drift else int(cho)

            on_drift = not on_drift
        return base
    
    def _generate_virtual(self, ncol, strength, drift_type="sudden"):
        n_cont, n_disc = map(int, ncol)
        base_coeff = rand(n_cont, 2)*10 #G[[k, theta]*n_cont]
        a = np.log2(strength)
        drift_coeff = np.array([[k, t+a] for (k, t) in base_coeff])

        base = {"time" : self.time}
        dumx = np.zeros(self._size) #rename
        attribute_cols = choice(np.arange(sum(ncol)), self.n_att, replace=False)
        self.attribute_columns = np.sort(attribute_cols)

        ### generate base continous columns
        for idx, (k, theta) in enumerate(base_coeff):
            gamma_dist = gamma(k, theta, self._size)
            base["C%02d"%idx] = gamma_dist
            if idx in attribute_cols:
                dumx += gamma_dist
        
        ### generate base discrete columns
        nsamples = randint(6, 30, n_disc)
        samples = [choice(np.arange(nsamples[i]), nsamples[i]//2, replace=False) for i in range(n_disc)]
        base_pdf = [rand(len(samples[i])) for i in range(n_disc)]
        drift_pdf = [rand(len(samples[i])) for i in range(n_disc)]
        for idx in range(n_disc):
            sample = self._discrete_sampling(samples[idx], base_pdf[idx], self._size)
            base["D%02d"%idx] = sample
            if int(idx+n_cont) in attribute_cols:
                dumx += sample
        gamma_dist /= len(attribute_cols)

        att_mean, att_std = np.mean(dumx), np.std(dumx)
        base = pd.DataFrame(base)

        threshold = att_mean + att_std/2

        if drift_type == "sudden":
            df = self._virtual_sudden(base=base, drift_coeff=drift_coeff,
            samples=samples, base_pdf = base_pdf, drift_pdf = drift_pdf)

        elif drift_type == "incremental":
            df = self._virtual_incremental(base=base, base_coeff=base_coeff, drift_coeff=drift_coeff,
            samples=samples, base_pdf = base_pdf, drift_pdf = drift_pdf)

        elif drift_type == "gradual":
            df = self._virtual_gradual(base=base, base_coeff=base_coeff, drift_coeff=drift_coeff,
            samples=samples, base_pdf = base_pdf, drift_pdf = drift_pdf)
        else:
            raise KeyError("There is no such type of drift [%s]"%drift_type)
        
        y = self._label_function(X = df.drop(columns="time").to_numpy(),
                                att_cols = attribute_cols,
                                threshold = threshold)
        df["y"] = y
        df["drift"] = self._drift_weight

        return df

    def generate(self, drift="real", time_format="mjd", drift_type="sudden", strength=10, noise=0.1):
        """
        Drift Generator
        drift : ["real", "virtual"]. drift의 종류 설정
        time : ["mjd", "datetime64"]. output의 time형식
        drift_Type : ["sudden", "incremental", "gradual"]. drift 발생 유형 설정
        strength : drift의 세기. default=10
        noise : data의 noise 비율
        """
        sanity = True
        while sanity:
            if drift == "real":
                df = self._generate_real(self._ncol, strength, drift_type)

            elif drift == "virtual":
                if drift_type != "sudden":
                    print("Generating virtual drift data may take a long time")
                df = self._generate_virtual(self._ncol, strength, drift_type)

            sanity = self._sanity_check(df)
            
        if time_format=="datetime64":
            df.time = Time(np.round(df.time.to_numpy(), 3), format="mjd").datetime64
        
        self.noise = noise
        if self.noise>0:
            for col in range(self._ncol[0]):
                df["C%02d"%col] = self._add_noise_cont(df["C%02d"%col].to_numpy())
            for col in range(self._ncol[1]):
                df["D%02d"%col] = self._add_noise_disc(df["D%02d"%col].to_numpy())
        
        self.df_info["strength"] = strength
        self.df_info["noise"] = noise
        self.df_info["drift"] = drift
        self.df_info["drift_type"] = drift_type
        self.df_info["att_cols"] = self.attribute_columns

        self.df = df.astype({"y":int})
        return self.df
    
    def save_to_csv(self, path, file_name = None, time_format="mjd", file_info=True):
        """
        path : csv를 저장할 디렉토리
        file_name : file의 이름. 미지정시 작성일자로 자동지정
        time_format : [mjd, datetime64, iso]
        file_info : df의 상세내용을 json으로 저장
        """
        print("Now saving...")
        if self.df is None:
            raise FileNotFoundError("Data yet been generated")
        if file_name is None:
            tp = self.df_info["prob_type"][0]+self.df_info["drift"][0] + self.df_info["drift_type"][0]
            file_name = "%s_%s"%(tp.upper(), datetime.now().strftime("%Y%m%d%H%M%S"))

        self.df.to_csv(path+"/%s.csv"%file_name)

        if file_info:
            self.df_info["prob_type"] = self.prob_type
            if time_format == "datetime64":
                self.df_info["time_range"] = str(Time(self.df_info["time_range"]).datetime64).tolist()
                self.df_info["drift_time"] = str(Time(self.df_info["drift_time"]).datetime64).tolist()
            elif time_format == "iso":
                self.df_info["time_range"] = str(Time(self.df_info["time_range"]).iso).tolist()
                self.df_info["drift_time"] = str(Time(self.df_info["drift_time"]).iso).tolist()
            else:
                try:
                    self.df_info["time_range"] = self.df_info["time_range"].tolist()
                    self.df_info["drift_time"] = self.df_info["drift_time"].tolist()
                except AttributeError:
                    pass
            self.df_info["att_cols"] = self.df_info["att_cols"].tolist()
            self.df_info["file"] = file_name
            self.df_info["date_generated"] = str(datetime.now())
            with open(path+"%s.json"%file_name, "w") as f:
                json.dump(self.df_info, f)
        print("Done")
        return 
    


from inspect import getmembers, isfunction, isclass, getfullargspec
from tabulate import tabulate
from artemis7.util.colorprint import Colorprint as cp

def list(class_):
    member_candidates = getmembers(class_, isfunction)
    ml = []
    print(cp.BOLD + "Functions" + cp.END)
    for mc in member_candidates:
        if class_.__name__ == mc[1].__module__:
            ml.append([mc[0], ", ".join(getfullargspec(mc[1]).args)])
    if not ml:
        print(cp.RED + "None" + cp.END)
    else:
        print(tabulate(ml))
    
    member_candidates = getmembers(class_, isclass)
    ml = []
    print(cp.BOLD + "Classes" + cp.END)
    for mc in member_candidates:
        if class_.__name__ == mc[1].__module__:
            ml.append([mc[0], ", ".join(getfullargspec(mc[1]).args)])
    if not ml:
        print(cp.RED + "None" + cp.END)
    else:
        print(tabulate(ml))
    return


import numpy as np
from scipy.linalg import sqrtm

def FID(act1, act2):
    """
    Calculate the FID between two images
    act1, act2 : 2D array.
    """
    mu1, sigma1 = np.mean(act1, axis=0), np.cov(act1, rowvar=False)
    mu2, sigma2 = np.mean(act2, axis=0), np.cov(act2, rowvar=False)
    ssdiff = np.sum((mu1 - mu2)**2.0)
    mat_dot = sigma1.dot(sigma2)
    covmean = sqrtm(mat_dot)
    if np.iscomplexobj(covmean):
        covmean = covmean.real
    fid = ssdiff + np.trace(sigma1 + sigma2 - 2.0 * covmean)
    return fid


import imageio
import os

class make_movie():
    def __init__(self, path, extension):
        """
        path : directory where input images exist
        extension : extension of images. (jpeg, png, ...)
        """
        self._file_list = [f for f in os.listdir(path) if f.endswith(extension)]
    def to_gif(self, **kwargs):
        """
        path : directory to save file
        output : name of save file
        fps : fps of gif
        """
        path = kwargs["path"]+"/" if "path" in kwargs.keys() else self.path+"/"
        if not os.path.exists(path):
            os.mkdir(path)
        with imageio.get_writer(path + kwargs["output"]+".gif", mode="I", duration=1/kwargs["fps"]) as writer:
            for filename in self._file_list:
                filename = os.path.join(path, filename)
                image = imageio.imread(filename)
                writer.append_data(image)
        return



from setuptools import setup, find_packages

setup(
    name = 'artemis7',
    version='0.0.3',
    description='Personal library to compute and draw',
    long_description = open("README.md").read(),
    long_description_content_type = "text/markdown",
    url ='https://github.com/dowoonlee/artemis.git',
    author='dwlee',
    author_email='dwlee717@gmail.com',
    license='artemis7',
    packages = ["artemis7", "artemis7.datagenerator", "artemis7.myplot", "artemis7.stats", "artemis7.util"],
    zip_safe=False,
    install_requires=[
        'numpy==1.22.3',
        'matplotlib==3.5.2',
        'astropy==5.1',
        'scipy==1.7.3',
        'pandas==1.4.3',
        'imageio==2.19.3'
    ]
)